/******************************************************************************
* Copyright (c) 2009 HarHid.labo
*
* Original code (ALivePDF Library) by Thibault Imbert.
*
* MPDF class supports to generate PDF files using Multi-Byte(Japanese).
* MPDF is based on the library of mbfpdf(php). 
* (mbfpdf.php version 1.0b : author is airwhite@airwhite.net)
*
* This MPDF License is same as following original ALivePDF Library's License. 
*
******************************************************************************/
/*
_________________            __________________________
___    |__  /__(_)__   _________  __ \__  __ \__  ____/
__  /| |_  /__  /__ | / /  _ \_  /_/ /_  / / /_  /_    
_  ___ |  / _  / __ |/ //  __/  ____/_  /_/ /_  __/
/_/  |_/_/  /_/  _____/ \___//_/     /_____/ /_/ 
..........................................................Multi-Byte Version


Copyright (c) 2008 Thibault Imbert

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package org.alivepdf.japanese.pdf {

  import org.alivepdf.pages.Page;
  import org.alivepdf.tools.sprintf;
  import flash.net.URLRequest;
  import flash.net.URLRequestHeader;
  import flash.net.URLRequestMethod;
  import flash.net.navigateToURL;
  import org.alivepdf.japanese.metrics.MBFontMetrics;
  import org.alivepdf.japanese.metrics.MBCMap;
  import org.alivepdf.japanese.fonts.MBFontFamily;
  import org.alivepdf.images.GIFImage;
  import org.alivepdf.images.JPEGImage;
  import org.alivepdf.images.PDFImage;
  import org.alivepdf.images.PNGImage;
  import org.alivepdf.saving.Method;

  import flash.utils.ByteArray;
  import flash.system.Capabilities;

  /**
  *
  * Using MBPDF class, you can generate the PDF document including the Japanese multiByte characters.<br/>
  * 日本語Multi-Byte対応を行います。<br/>
  */
  public class MBPDF extends PDF_mdf {

    //--- charSet ------
    public static const SHIFT_JIS:String = 'shift_jis';
    public static const EUC_JP:String    = 'euc-JP';
    public static const UTF_8:String     = 'utf-8';
    public static const UTF_16:String    = 'utf-16';

    //--- MBCMap ------
    public static const CMap_BIG5:String   = "BIG5";
    public static const CMap_GB:String     = "GB";
    public static const CMap_SJIS:String   = "SJIS";
    public static const CMap_EUC_JP:String = "EUC_JP";
    public static const CMap_UniJIS_UTF16_H:String = "UniJIS_UTF16_H";
    public static const CMap_UniJIS_UTF16_V:String = "UniJIS_UTF16_V";
    public static const CMap_UniJIS_UTF8_H:String  = "UniJIS_UTF8_H";
    public static const CMap_UniJIS_UTF8_V:String  = "UniJIS_UTF8_V";

    //--- Version  ------
    protected static const MBPDF_VERSION:String = '0.0.1.c';

    /**
    * The MBPDF class represents a PDF document including the Japanese multiByte characters.
    * @author HarHid  (PDF class's author:Thibault Imbert)
    * @example
    * This example shows how to create a valid PDF document :
    * <div class="listing">
    * <pre>
    * var myPDF:MBPDF = new MBPDF ( Orientation.PORTRAIT, Unit.MM, Size.A4 );
    * </pre>
    * </div>
    */
    public function MBPDF ( orientation:String='Portrait', unit:String='Mm', pageSize:Object=null, rotation:int=0 ) {
      if(ALIVEPDF_VERSION != '0.1.4.8'){
        throw new Error ('AlivePDF Version must be 0.1.4.8 !!');
      }
    }

    /**
    * Add a Multi-Byte font family.
    *
    * @param family Multi-Byte font family name, see MBFontFamily.
    * @param enc Multi-Byte encoding name, value can be SHIFT_JIS, EUC_JP, UTF_8, UTF_16.
    * @example
    * This example shows how to add Multi-Byte font family and to set this font family.
    * <div class="listing">
    * <pre>
    * myPDF.addMBFont(MBFontFamily.MINCHO ,'SJIS');
    * myPDF.setFont(MBFontFamily.MINCHO,'',50);
    * myPDF.addCell(50,10,'アイウエオ',1,1);
    * </pre>
    * </div>
    * @throw Error invalid family parameter
    * @throw Error invalid enc parameter
    * @see org.alivepdf.japanese.font.MBFontFamily#BIG5
    * @see org.alivepdf.japanese.font.MBFontFamily#GB
    * @see org.alivepdf.japanese.font.MBFontFamily#KOZMIN
    * @see org.alivepdf.japanese.font.MBFontFamily#GOTHIC
    * @see org.alivepdf.japanese.font.MBFontFamily#PGOTHIC
    * @see org.alivepdf.japanese.font.MBFontFamily#UIGOTHIC
    * @see org.alivepdf.japanese.font.MBFontFamily#MINCHO
    * @see org.alivepdf.japanese.font.MBFontFamily#PMINCHO
    * @see #SHIFT_JIS
    * @see #EUC_JP
    * @see #UTF_8
    * @see #UTF_16
    */
    public function addMBFont(family:String = '',enc:String = ''):void {
trace("addMBFont family="+family+" enc="+enc);
      var gt:Class = MBFontMetrics;  //multi-bytes font-metrics def
      var gc:Class = MBCMap;         //Multi-bytes Cmap Information

      if (enc == '' || gc[enc] == null) {
        throw new Error("AddMBFont: ERROR Encoding [" + enc + "] Undefine.");
      }
      if (gt[family]!=null) {
        var ut:Number = gt[family]['ut'];    // UnderlineThickness
        var up:Number = gt[family]['up'];    // UnderlinePosition
        var cw:Object = gt[family]['cw'];    // Character Width
        var cm:String = gc[enc]['CMap'];     // CMap name
        var od:String = gc[enc]['Ordering']; 
        var sp:int = gc[enc]['Supplement'];
        var registry:Object = {ordering:od, supplement:sp};

        this.AddCIDFont(family, ''   ,family               ,cw,cm,registry,ut,up);
        this.AddCIDFont(family, 'B'  ,family+",Bold"       ,cw,cm,registry,ut,up);
        this.AddCIDFont(family, 'I'  ,family+",Italic"     ,cw,cm,registry,ut,up);
        this.AddCIDFont(family, 'BI' ,family+",BoldItalic" ,cw,cm,registry,ut,up);
      } 
      else {
        throw new Error("AddMBFont: ERROR FontName [" + family + "] Undefine.");
      }
    }
    /**
    * @private
    */
    protected function AddCIDFont(family:String, style:String , name:String, cw:Object,cMap:String,registry:Object,ut:Number,up:Number):void {
      var i:int = getNumImages(fonts)+1;
      fontkey = family.toLowerCase() + style.toUpperCase();
      this.fonts[fontkey] =
        { i:i,type:'Type0',name:name,up:up,ut:ut,cw:cw,CMap:cMap,registry:registry};
    }

    /**
    * @private
    */
    override protected function getStringWidth(content:String ):Number {
      if(this.currentFont.type == 'Type0')
        return this.getMBStringWidth(content);
      else
        return super.getStringWidth(content);
    }

    /**
    * @private
    */
    protected function getMBStringWidth( content:String ):Number{
      //Multi-byte(UTF-8) version of GetStringWidth()
      //ActionScript3 Internal Character Set is UTF-8.

      var l:int = 0;
      cw = this.currentFont.cw;
      var japanese:Boolean = (this.currentFont.registry.ordering == 'Japan1');
 
      var nb:Number = content.length;
      var i:int = 0;

      while(i < nb){
        var c:String = content.charAt(i);
        var ascii:Boolean = checkAscii(c);
        var hwkana:Boolean = japanese && checkHalfKana(c);
        if(ascii){
          l += cw[c];
          i++;
        }
        else{
          l += hwkana ? 500 : 1000;
          i++;
        }
      }
      return l * this.fontSize/1000;
    }


    /**
    * Add a cell with some text to the current page
    * It is acceptable to write the Japanese multi-byte characters.
    *
    * @param width Cell width
    * @param height Cell height
    * @param text Text to add into the cell
    * @param ln Sets the new position after cell is drawn, default value is 0
    * @param align Lets you center or align the text into the cell
    * @param fill Lets you specify if the cell is colored (1) or transparent (0)
    * @param link Any http link, like http://www.mylink.com
    * @return Page
    * @example
    * This example shows how to write some text within a cell :
    * <div class="listing">
    * <pre>
    * myPDF.setFont(FontFamily.HELVETICA, 'B', 12);
    * myPDF.addMBFont(MBFontFamily.GOTHIC ,'SJIS');
    * myPDF.setFont(MBFontFamily.GOTHIC, 'B', 50);
    * myPDF.textStyle ( new RGBColor ( 0x990000 ) );
    * myPDF.addCell(50,10,'何かのテキストをセルに入れる Some text into a cell !',1,1);
    * </pre>
    * </div>
    * This example shows how to write some clikable text within a cell :
    * <div class="listing">
    * <pre>
    * myPDF.setFont(FontFamily.HELVETICA, 'B', 12);
    * myPDF.addMBFont(MBFontFamily.GOTHIC ,'SJIS');
    * myPDF.setFont(MBFontFamily.GOTHIC, 'B', 50);
    * myPDF.textStyle ( new RGBColor ( 0x990000 ) );
    * myPDF.addCell(50,10,'クリックできます A clikable cell !', 1, 1, null, 0, "http://www.alivepdf.org");
    * </pre>
    * </div>
    */
    override public function addCell( width:Number=0, height:Number=0, text:String='', border:*=0, ln:Number=0, align:String='', fill:Number=0, link:String='' ):void {

      var k:Number = this.k;

      if( this.currentY + height > this.pageBreakTrigger && !this.inFooter && this.acceptPageBreak() ){
        //Automatic page break
        var x:Number = this.currentX;
        ws=this.ws;
        if(ws>0){
          this.ws=0;
          this.write('0 Tw');
        }
        this.addPage( new Page ( this.currentOrientation, this.defaultUnit, this.defaultSize ,currentPage.rotation ) );
        this.currentX = x;
        if(ws>0){
          this.ws=ws;
          this.write(sprintf('%.3f Tw',ws*k));
        }
      }

      if ( width==0 ) width = currentPage.w-this.rMargin-this.currentX;
	
      var s:String = new String();
      var op:String;

      if( fill == 1 || border == 1 ){
        if ( fill == 1 ) op = ( border == 1 ) ? 'B' : 'f';
        else op = 'S';
				
        s = sprintf('%.2f %.2f %.2f %.2f re %s ',this.currentX*k,(currentPage.h-this.currentY)*k,width*k,-height*k,op);
      }

      if ( border is String ){
        currentX = this.currentX;
        currentY = this.currentY;

        var tmpBorder:String = String ( border );

        if( tmpBorder.indexOf('L') != -1 ) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-currentY)*k,currentX*k,(currentPage.h-(currentY+height))*k);
        if( tmpBorder.indexOf ('T') != -1) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-currentY)*k,(currentX+width)*k,(currentPage.h-currentY)*k);
        if( tmpBorder.indexOf ('R') != -1) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',(currentX+width)*k,(currentPage.h-currentY)*k,(currentX+width)*k,(currentPage.h-(currentY+height))*k);
        if( tmpBorder.indexOf ('B') != -1 ) s+=sprintf('%.2f %.2f m %.2f %.2f l S ',currentX*k,(currentPage.h-(currentY+height))*k,(currentX+width)*k,(currentPage.h-(currentY+height))*k);
      }

      if ( text !== '' ){
			
        var dx:Number;
        if ( align=='R' ) dx = width-this.cMargin-this.getStringWidth(text);
        else if( align=='C' ) dx = (width-this.getStringWidth(text))/2;
        else dx = this.cMargin;

        if(this.colorFlag) s+='q '+this.addTextColor+' ';

        //Convert charSet of output string (UTF-8 => specified charset) and convert <HexDump>.
        s += sprintf('BT %.2f %.2f Td <%s> Tj ET',(this.currentX+dx)*k,(currentPage.h-(this.currentY+.5*height+.3*this.fontSize))*k,charSetConv(text));

        if(this.underline) s+=' '+doUnderline(this.currentX+dx,this.currentY+.5*height+.3*this.fontSize,text);
        if(this.colorFlag) s+=' Q';
        if( link ) this.addLink (this.currentX+dx,this.currentY+.5*height-.5*this.fontSize,this.getStringWidth(text),this.fontSize, link);

      }

      if ( s ) this.write(s);

      this.lasth = currentPage.h;

      if( ln > 0 ){
        //Go to next line
        this.currentY += height;
        if( ln ==1) this.currentX = this.lMargin;

      } 
      else
        this.currentX += width;
    } // end of the "addCell()" method

    /**
    * Add a multicell with some text to the current page
    * It is acceptable to write the multi-byte(Japanese) characters.
    *
    * @param width Cell width
    * @param height Cell height
    * @param text Text to add into the cell
    * @param border Lets you specify if a border should be drawn around the cell
    * @param align Lets you center or align the text into the cell, values can be L (left align), C (centered), R (right align), J (justified) default value
    * @param filled Lets you specify if the cell is colored (1) or transparent (0)
    * @return Page
    * @example
    * This example shows how to write a table made of text cells :
    * <div class="listing">
    * <pre>
    * myPDF.setFont( FontFamily.COURIER, Style.BOLD, 14 );
    * mypdf.addMBFont(MBFontFamily.GOTHIC ,'SJIS');
    * myPDF.setFont(MBFontFamily.GOTHIC, 'B', 50);
    * myPDF.textStyle ( new RGBColor ( 0x990000 ) );
    * myPDF.addMultiCell ( 70, 24, "マルチセル A multicell :)", 1);
    * myPDF.addMultiCell ( 70, 24, "マルチセル A multicell :)", 1);
    * </pre>
    * </div>
    */

    override public function addMultiCell(width:Number, height:Number, text:String, border:*=0, align:String='L', filled:int=0):void {
       if(this.currentFont.type == 'Type0')
         this.addMBMultiCell(width,height,text,border,align,filled);
       else
         super.addMultiCell(width,height,text,border,align,filled); 
    }

    /**
    * @private
    */
    protected function addMBMultiCell(width:Number, height:Number, text:String, border:*=0, align:String='L', filled:int=0):void {
      //Multi-byte version of addMultiCell()
      cw = this.currentFont.cw;
      var japanese:Boolean = (this.currentFont.registry.ordering == 'Japan1');
      if(width == 0)
        width = currentPage.w - this.rMargin - this.currentX;
      var wmax:Number = (width - 2 * this.cMargin)*1000 / this.fontSize;
      var s:String = findAndReplace("\r",'',text);
      nb = s.length;

      if(nb > 0 && s.charAt(nb-1) == "\n")
        nb--;

      var b:* = 0;
      if(border) {
        if(border == 1) {
          border = 'LTRB';
          b = 'LRT';
          b2 = 'LR';
        }
        else {
          b2 = '';
          if(border.indexOf('L')!= -1)
            b2 += 'L';
          if(border.indexOf('R')!= -1)
            b2 += 'R';
          b = border.indexOf('T')!= -1 ? b2+'T' : b2;
        }
      }

      var sep:int = -1;
      var i:int = 0;
      var j:int = 0;
      var l:int = 0;
      var ns:int = 0;
      var nl:int = 1;
      var ls:int;
      var c:String;
      var cwAux:int = 0;    //*** harhid
      while(i < nb){
        //Get next character
        c = s.charAt(i);

        var prev_ascii:Boolean = ascii;

        var ascii:Boolean = checkAscii(c);
        var hwkana:Boolean = japanese && checkHalfKana(c);

        if(c == "\n"){
          //Explicit line break
          if(this.ws > 0){
            this.ws = 0;
            this.write('0 Tw');
          }
          this.addCell(width,height,s.substr(j,i-j),b,2,align,filled);
          i++;
          sep = -1;
          j = i;
          l = 0;
          ns = 0;
          nl++;
          if(border && nl == 2)
            b = b2;
          continue;
        }
        /*** if current is not ASCII and previous is ASCII **/
        if(!(ascii && prev_ascii) && i != j){
          sep = i;
          ls = l;
        }
        else if(c == ' '){
          sep = i;
          ls = l;
          ns++;
        }
        // TB0

        cwAux = cw[c] as int;
        if(cwAux == 0) cwAux = 580;
        l += ascii ? cwAux : hwkana ? 500 : 1000;

        if(l > wmax){
          //Automatic line break
          if(sep == -1){
            if(i == j)
              i += 1;  
            if(this.ws > 0){
              this.ws = 0;
              this.write('0 Tw');
            }
            this.addCell(width,height,s.substr(j,i-j),b,2,align,filled);
          }
          else {
            if(align == 'J'){
              if(s.charAt(sep) == ' ')
                ns--;
              if(s.charAt(i-1) == ' '){
                ns--;
                ls -= cw[' '];
              }
              this.ws = (ns>0) ? (wmax-ls)/1000*this.fontSize/ns : 0;
              this.write(sprintf('%.3f Tw',this.ws*this.k));
            }
            this.addCell(width,height,s.substr(j,sep-j),b,2,align,filled);
            i = (s.charAt(sep) == ' ') ? sep+1 : sep;
          }
          sep = -1;
          j = i;
          l = 0;
          ns = 0;
          nl++;
          if(border && nl == 2)
            b = b2;
        }
        else
          i += 1;

      }
      //Last chunk
      if(this.ws > 0){
        this.ws = 0;
        this.write('0 Tw');
      }
      if(border && border.indexOf ('B')!= -1)
        b += 'B';
      this.addCell(width,height,s.substr(j,i-j),b,2,align,filled);
      this.currentX = this.lMargin;
    }

   /**
    * Add a cell with some text to the current page.
    * It is acceptable to write the multi-byte(Japanese) characters.
    *
    * @param width Cell width
    * @param height Cell height
    * @param text Text to add into the cell
    * @param ln Sets the new position after cell is drawn, default value is 0
    * @param align Lets you center or align the text into the cell
    * @param fill Lets you specify if the cell is colored (1) or transparent (0)
    * @param link Any http link, like http://www.mylink.com
    * @return Page
    * @example
    * This example shows how to write some text within a cell :
    * <div class="listing">
    * <pre>
    * myPDF.setFont( FontFamily.COURIER, Style.BOLD, 14 );
    * myPDF.addMBFont(MBFontFamily.MINCHO ,'SJIS');
    * myPDF.setFont(MBFontFamily.MINCHO,'B',50);
    * myPDF.textStyle ( new RGBColor ( 0x990000 ) );
    * myPDF.addCell(50,10,'なにか文字をセルに書き込む !',1,1);
    * </pre>
    * </div>
    * This example shows how to write some clikable text within a cell :
    * <div class="listing">
    * <pre>
    * myPDF.setFont( FontFamily.COURIER, Style.BOLD, 14 );
    * myPDF.addMBFont(MBFontFamily.MINCHO ,'SJIS');
    * myPDF.setFont(MBFontFamily.MINCHO,'B',50);
    * myPDF.textStyle ( new RGBColor ( 0x990000 ) );
    * myPDF.addCell(50,10,'クリックできます。A clikable cell !', 1, 1, null, 0, "http://www.alivepdf.org");
    * </pre>
    * </div>
    */

    override public function addText ( text:String, x:Number=0, y:Number=0 ):void {
      if(this.currentFont.type == 'Type0')
        this.mbAddText(text,x,y);
      else
        super.addText(text,x,y);
    }

    /**
    * @private
    */
    protected function mbAddText ( text:String, x:Number=0, y:Number=0 ):void {
      var s:String = sprintf('BT %.2f %.2f Td <%s> Tj ET',x*k, (currentPage.h-y)*k, charSetConv(text));
      if (underline && text !='') s += ' '+doUnderline(x,y,text);
      if (colorFlag) s = 'q ' + addTextColor + ' ' + s +' Q';
      write(s);
    }


    /**
    * Lets you write some text. 
    * It is acceptable to write the Japanese multi-byte characters.
    *
    * @param lineHeight Line height, lets you specify height between each lines
    * @param text Text to write, to put a line break just add a \n in the text string
    * @param link Any link, like http://www.mylink.com, will open te browser when clicked
    * @throw Error text parameter value is null.
    * @example
    * This example shows how to add some text to the current page :
    * <div class="listing">
    * <pre>
    * myPDF.setFont( FontFamily.COURIER, Style.BOLD, 14 );
    * myPDF.addMBFont(MBFontFamily.GOTHIC ,'SJIS');
    * myPDF.setFont(MBFontFamily.GOTHIC, 'B', 50);
    * myPDF.writeText ( 5, "愛は美しい, Love is beautiful.");
    * </pre>
    * </div>
    * This example shows how to add some text with a clickable link :
    * <div class="listing">
    * <pre>
    * myPDF.setFont( FontFamily.COURIER, Style.BOLD, 14 );
    * myPDF.addMBFont(MBFontFamily.GOTHIC ,'SJIS');
    * myPDF.setFont(MBFontFamily.GOTHIC, 'B', 50);
    * myPDF.writeText ( 5, "愛は美しい, Love is beautiful., "http://www.google.fr");
    * </pre>
    * </div>
    */

    override public function writeText(lineHeight:Number, text:String, link:String='' ):void {
      if(text == null) throw new Error('Input text is null data. Check your text data.');
      if(this.currentFont.type == 'Type0')
        this.mbWriteText(lineHeight, text,link);
      else
        super.writeText(lineHeight, text,link);
    }
   
    /**
    * @private
    */
    protected function mbWriteText( lineHeight:Number, text:String, link:String='' ):void {

      var cw:Object = this.currentFont.cw;
      var japanese:Boolean = (this.currentFont.registry.ordering == 'Japan1');
      var w:Number = currentPage.w - this.rMargin - this.currentX;

      var wmax:Number = (w - 2 * this.cMargin)*1000/this.fontSize;
      var s:String = findAndReplace("\r",'',text);
      var nb:int = s.length;
      var sep:int = -1;
      var i:int = 0;
      var j:int = 0;
      var l:int = 0;
      var nl:int =1;
      var c:String;
      while(i < nb){
        //Get next character
        c = s.charAt(i);
        var ascii:Boolean = checkAscii(c);
        var hwkana:Boolean = japanese && checkHalfKana(c);

        if(c == "\n"){
          //Explicit line break
          this.addCell(w,lineHeight,s.substr(j,i-j),0,2,'',0,link);
          i++;
          sep = -1;
          j = i;
          l = 0;
          if(nl == 1){
            this.currentX = this.lMargin;
            w = currentPage.w - this.rMargin - this.currentX;
            wmax = (w - 2*this.cMargin)*1000/this.fontSize;
          }
          nl++;
          continue;
        }
        if(!ascii || c == ' ')
          sep = i;
        l += ascii ? cw[c] : hwkana ? 500 : 1000;
        if(l > wmax){
          //Automatic line break
          if(sep == -1 || i == j){
            if(this.currentX > this.lMargin){
              //Move to next line
              this.currentX = this.lMargin;
              this.currentY += currentPage.h;
              w = currentPage.w - this.rMargin- this.currentX;
              wmax = (w-2*this.cMargin)*1000/this.fontSize;
              i++;
              nl++;
              continue;
            }
            if(i == j){
              i += 1;
              //i += ascii ? 1 : 2;
            }
            this.addCell(w,lineHeight,s.substr(j,i-j),0,2,'',0,link);
          }
          else{
            this.addCell(w,lineHeight,s.substr(j,sep-j),0,2,'',0,link);
            i = (s.charAt(sep) == ' ') ? sep+1 : sep;
          }
          sep=-1;
          j = i;
          l = 0;
          if(nl == 1){
            this.currentX = this.lMargin;
            w = currentPage.w - this.rMargin - this.currentX;
            wmax = (w-2*this.cMargin)*1000/this.fontSize;
          }
          nl++;
        }
        else{
          i += 1;
        }
      }
      //Last chunk
      if(i != j){
        this.addCell(l/1000*this.fontSize,lineHeight,s.substr(j,i-j),0,0,'',0,link);
      }
    }

    /**
    * @private
    */
    override protected function insertFonts():void {
      var nf:int = this.n;
      for(var diff:String in this.diffs) {
       //Encodings
       this.newObj();
       this.write('<</Type /Encoding /BaseEncoding /WinAnsiEncoding /Differences ['+diff+']>>');
       this.write('endobj');
      }

      for(var k:String in this.fonts) {
        //Font objects
        this.newObj();
        var font:Object = this.fonts[k];
        font.n=this.n;
        this.write('<</Type /Font');
        if(font.type=='Type0')
          this._puttype0(font);
        else {
          var name:String = font.name;
          this.write('/BaseFont /'+name);
          if(font.type=='core') {
            //Standard font
            this.write('/Subtype /Type1');
            if(name!='Symbol' && name!='ZapfDingbats')
              this.write('/Encoding /WinAnsiEncoding');
          }
          else {
            //Additional font
            this.write('/Subtype /'+font.type);
            this.write('/FirstChar 32');
            this.write('/LastChar 255');
            this.write('/Widths '+(this.n+1)+' 0 R');
            this.write('/FontDescriptor '+(this.n+2)+' 0 R');
            if(font.enc) {
              if(font.diff != null)
                this.write('/Encoding '+(nf+font.diff)+' 0 R');
              else
                this.write('/Encoding /WinAnsiEncoding');
            }
          }
          this.write('>>');
          this.write('endobj');
          if(font.type!='core'){
            //Widths
            this.newObj();
            var cw:Object = font.cw;
            var s:String ='[';
            for(var i:int =32; i<=255; i++)
              s += cw[String.fromCharCode(i)] + ' ';
            this.write(s+']');
            this.write('endobj');
            //Descriptor
            this.newObj();
            s = '<</Type /FontDescriptor /FontName /'+name;
            for(var q:String in font.desc) 
              s += ' /'+q+' '+font.desc[q];
            var file:Object = font.file;
            if(file != null)
              s += ' /FontFile'+(font.type=='Type1' ? '' : '2')+' '+this.fontFiles[file].n +' 0 R';
            this.write(s+'>>');
            this.write('endobj');
          }
        }
      }
    }

    /**
    * @private
    */
    protected function _puttype0(font:Object):void {
      //Type0    
      var name:String = font.name;
      var cMap:String = font.CMap;
      var registry:Object = font.registry;
      var cw:Object = font.cw;

      this.write('/Subtype /Type0');
      this.write('/BaseFont /' + name + '-' + cMap);
      this.write('/Encoding /' + cMap);
      this.write('/DescendantFonts [' + (this.n+1) + ' 0 R]');
      this.write('>>');
      this.write('endobj');
      //CIDFont
      this.newObj();
      this.write('<</Type /Font');
      this.write('/Subtype /CIDFontType0');
      this.write('/BaseFont /' + name);
      this.write('/CIDSystemInfo <</Registry (Adobe) /Ordering ('+ registry.ordering + ') /Supplement ' + registry.supplement + '>>');
      this.write('/FontDescriptor ' + (this.n+1) + ' 0 R');
      var s:String = '/W [1 [';
      for (var q:String in cw )
        s += cw[q] + ' ';
      this.write(s + ']');
      if(registry.ordering == 'Japan1')
        this.write(' 231 325 500 631 [500] 326 389 500');
      this.write(']');
      this.write('>>');
      this.write('endobj');
      //Font descriptor
      this.newObj();
      this.write('<</Type /FontDescriptor');
      this.write('/FontName /' + name);
      this.write('/Flags 6');
      this.write('/FontBBox [0 0 1000 1000]');
      this.write('/ItalicAngle 0');
      this.write('/Ascent 1000');
      this.write('/Descent 0');
      this.write('/CapHeight 1000');
      this.write('/StemV 10');
      this.write('>>');
      this.write('endobj');
    }

    /**
    * character.charCodeAt()の値をもとに、ASCII文字であるか否かを判定します.
    * 以下の条件の場合、ASCII文字と判定します.<br/>
    * <div class="listing">
    * <pre>
    * cn < 128 or cn == 0x00a5('\') or cn == 0x203e('~')
    * ここで、cn = character.charCodeAt()。
    * </pre>
    * </div>
    * 
    */
    protected function checkAscii(character:String):Boolean {
      //128 = x80  this means ASCII character
      // 0x00a5(unicode) <==> \ , 0x203e(unicode) <==> ~
      var cn:uint = character.charCodeAt();
      var ascii:Boolean = (cn < 128 || cn == 0x00a5 || cn == 0x203e );
      return ascii;
    }

    /**
    * character.charCodeAt()の値をもとに、半角カナ文字であるか否かを判定します.
    * 以下の条件の場合、半角カナ文字と判定します.<br/>
    * <div class="listing">
    * <pre>
    * character.charCodeAt()が、0xff61から0xff9fの場合。
    * </pre>
    * </div>
    * 
    */
    protected function checkHalfKana(s:String):Boolean {
      //Unicode FF61 - FF9F  means half wide Kana.
      var cn:uint = s.charCodeAt();
      var hwkana:Boolean = (0xff60 < cn) && (cn < 0xffa0);
      return hwkana;
    }

    /**
    * @private
    */
    protected function charSetConv(s:String):String {
      /* Change the character-encoding          */
      var buffer:ByteArray = new ByteArray();
      buffer.writeMultiByte(s, this.charSet);
      buffer.position = 0;
      var bytes : String = this.readByte_HexString(buffer);
      if(this.charSet == MBPDF.UTF_16 && bytes.length == 2) bytes = '00'+ bytes;
      return bytes;
    }

    /**
    * @private
    */
    protected function readByte_HexString(buffer:ByteArray,l:Number=0):String {
      buffer.position = 0;
      var bytes : String = "";
      if(l==0){l=buffer.length;}
      for (var i : int = 0; i < buffer.length && i < l; i++) {
        var a:int = buffer.readByte();
        var u:uint = (a < 0)? 0xff + a + 1: a;
        bytes += uint2hex(u);
       }
      return bytes;
    }
    /**
    * @private
    */
    protected const HEX : String = "0123456789ABCDEF";
    /**
    * コード値(uint)を16進表示に変換する.
    * @parameter d  コード値
    * @return  16進表示の文字列
    * @throw Error 入力のコード値が１バイト分を超える(0xff超)
    */
    protected function uint2hex(d:uint):String {
      if(d >>> 8) throw new Error('MBPDF#uint2hex() : error d='+d);
      return HEX.charAt( d  >>> 4 & 0xf) + HEX.charAt( d & 0xf);
    }

    /**
    * @private
    */
    protected var _charSet:String = UTF_16;
    /**
    * This property is used to convert the encoding of Japanese Multi-Byte characters.
    * <br/>
    * The charSet value can be one of following values: 
    * <br/>
    * <pre>
    * MBPDF.SHIFT_JIS
    * MBPDF.EUC_JP
    * MBPDF.UTF_8
    * MBPDF.UTF_16
    * </pre>
    */
    public function get charSet():String {
      return _charSet;
    }
    public function set charSet(value:String):void {
      _charSet = value;
    }
  }
}

